# BACKEND
class InvalidSigmAIQBackend(Exception):
    """
    Exception for the :class:`Generic SigmAIQ Backend Factory
    <sigmaiq.sigma_backend_factory>` factory.

    To be used when the selected backend in the factory is invalid.
    """


class InvalidOutputFormat(Exception):
    """
    Exception for the :class:`Generic SigmAIQ Backend Class
    <sigmaiq.sigma_backend_class>` class.

    To be used when the specified output_format for the class is invalid.
    """


# PIPELINE
class InvalidSigmAIQPipeline(Exception):
    """
    Exception for the :class:`Generic SigmAIQ Pipeline Factory
    <sigmaiq.sigma_pipeline_factory>` factory.

    To be used when the provided processing_pipeline in the factory is invalid.
    """


class InvalidCustomFieldMapping(Exception):
    """
    Exception for the :class:`Generic SigmAIQ Pipeline Factory
    <sigmaiq.sigma_pipeline_factory.SigmAIQPipeline>` factory.

    To be used when the provided field mapping to the classmethod `from_fieldmap()`
    in the factory is invalid or empty.
    """


# RULEBUILDER
class InvalidSigmaRuleFilepath(Exception):
    """For use when the Sigma Rule Filepath in globals is not set or invalid"""


class InvalidIOCField(Exception):
    """For use when ioc_field or ioc_value isn't in an IOC dictionary used for creating detections"""



import asyncio
import json
from typing import Optional, Type, Union

from langchain.prompts import ChatPromptTemplate
from langchain.schema.language_model import BaseLanguageModel
from langchain.schema.output_parser import StrOutputParser
from langchain.schema.runnable import RunnablePassthrough
from langchain.tools import BaseTool
from pydantic import BaseModel, Extra, Field

from detectiq.sigmaiq.sigmaiq_backend_factory import AVAILABLE_BACKENDS


class QueryToSigmaRuleInput(BaseModel):
    """Input for QueryToSigmaRule tool, which converts a backend query to a Sigma Rule, and
    uses SigmAIQ backend factory for validation.
    """

    query: Union[str, dict] = Field(
        default=None,
        description="""A query string for a backend, which should be converted to a Sigma Rule YAML string.""",
    )
    backend: str = Field(
        default=None,
        description="""The backend that the query is for, and what should be used for validation. Backend options their descriptions are as 
        follows:\n"""
        + f"{json.dumps(AVAILABLE_BACKENDS, indent=2)}",
    )


class QueryToSigmaRuleTool(BaseTool):
    """Class for converting a backend query to a Sigma Rule"""

    name: str = "query_to_sigma_rule"
    args_schema: Type[BaseModel] = QueryToSigmaRuleInput
    description: str = """
Use this tool to take an already existing query for a backend and convert it to a Sigma Rule.
Use the translate_sigma_rule tool to take the created Sigma Rule and convert it back to a query to 
determine if the queries are the same and the conversion is successful. Fix any errors if necessary.
The input must be a query string for a backend, and the backend must be specified.
The output is a Sigma Rule YAML string, or an error message if the conversion fails.
"""
    # return_direct = True  # We don't need an agent LLM to think about the output, it is what it is.
    llm: BaseLanguageModel
    verbose: bool = False

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid

    def _run(self, query: Optional[str] = None, backend: Optional[str] = None) -> str:
        return asyncio.run(self._arun(query, backend))

    async def _arun(self, query: Optional[str] = None, backend: Optional[str] = None) -> str:
        template = """You are a cybersecurity detection engineering assistant bot specializing in Sigma Rule creation.
You are assisting a user in taking a query for a security/SIEM product, and converting it to a Sigma Rule.
The backend is used to validate the query and ensure it is compatible with the backend.
The created Sigma Rule should be in YAML format and use the official Sigma schema.  The detection field
can contain multiple 'selection' identifiers and multiple 'filter' identifiers as needed, 
which can be used in the condition field to select criteria and filter out criteria respectively.  
The fields should be Sysmon field names if possible, or Windows Event Log field names if possible.  

-----------

Sigma Rule Schema:

title
id [optional]
related [optional]
   - id {{rule-id}}
     type {{type-identifier}}
status [optional]
description [optional]
references [optional]
author [optional]
date [optional]
modified [optional]
tags [optional]
logsource
   category [optional]
   product [optional]
   service [optional]
   definition [optional]
   ...
detection
   {{search-identifier}} [optional]
      {{string-list}} [optional]
      {{map-list}} [optional]
      {{field: valu}}> [optional]
   ... # Multiple search identifiers can be specified as needed and used in the condition
   condition
fields [optional]
falsepositives [optional]
level [optional]:

-----------

User's Query: {query}
Backend: {backend}
"""

        prompt = ChatPromptTemplate.from_template(template)
        chain = (
            {"query": RunnablePassthrough(), "backend": RunnablePassthrough()} | prompt | self.llm | StrOutputParser()
        )
        return await chain.ainvoke({"query": query, "backend": backend})



from typing import Callable, Dict, List, Optional, Union
from uuid import uuid4

from sigma.pipelines.azuremonitor import azure_monitor_pipeline

## Carbon Black
from sigma.pipelines.carbonblack import (
    CarbonBlack_pipeline,
    CarbonBlackResponse_pipeline,
)

## Cortex XDR
from sigma.pipelines.cortexxdr import CortexXDR_pipeline

## Crowdstrike
from sigma.pipelines.crowdstrike import (
    crowdstrike_falcon_pipeline,
    crowdstrike_fdr_pipeline,
)

## Elasticsearch
from sigma.pipelines.elasticsearch import (
    ecs_kubernetes,
    ecs_windows,
    ecs_windows_old,
    ecs_zeek_beats,
    ecs_zeek_corelight,
    zeek_raw,
)

## Loki
from sigma.pipelines.loki import (
    loki_grafana_logfmt,
    loki_okta_system_log,
    loki_promtail_sysmon,
)

## Microsoft
from sigma.pipelines.microsoftxdr import microsoft_xdr_pipeline

# Netwitness
from sigma.pipelines.netwitness import netwitness_windows_pipeline

## QRadar
from sigma.pipelines.QRadarAQL import (
    QRadarAQL_fields_pipeline,
    QRadarAQL_payload_pipeline,
)

## SecOps
from sigma.pipelines.secops import secops_udm_pipeline
from sigma.pipelines.sentinelasim import sentinel_asim_pipeline

## SentinelOne
from sigma.pipelines.sentinelone import sentinelone_pipeline

## Splunk
from sigma.pipelines.splunk import (
    splunk_cim_data_model,
    splunk_windows_pipeline,
    splunk_windows_sysmon_acceleration_keywords,
)

## STIX
from sigma.pipelines.stix import stix_2_0, stix_shifter

## Sysmon
from sigma.pipelines.sysmon import sysmon_pipeline

## Windows
from sigma.pipelines.windows import windows_audit_pipeline, windows_logsource_pipeline
from sigma.processing.pipeline import ProcessingItem, ProcessingPipeline
from sigma.processing.resolver import ProcessingPipelineResolver
from sigma.processing.transformations import FieldMappingTransformation

from detectiq.sigmaiq.exceptions import (
    InvalidCustomFieldMapping,
    InvalidSigmAIQPipeline,
)

## AIQ
from detectiq.sigmaiq.pipelines.splunk_windows_audit import splunk_wineventlog_pipeline

#############
# PIPELINES #
#############
## InsightIDR
# RS uncommented this line after Stephen uncomment corresponding line in pyproject.toml
# from sigma.pipelines.insight_idr import insight_idr_pipeline

AVAILABLE_PIPELINES = {  # AIQ Custom
    "splunk_wineventlog": {
        "description": "SigmAIQ Custom combined windows_audit and splunk_windows pipelines to convert Sysmon fields to "
        "Windows Event Log fields for Splunk searches",
        "pipeline": splunk_wineventlog_pipeline(),
        "display_name": "Splunk WinEventLog",
    },  # CarbonBlack
    "carbonblack": {
        "description": "Uses Carbon Black EDR field mappings",
        "pipeline": CarbonBlackResponse_pipeline(),
        "display_name": "CB",
    },  # Cortex XDR, Palo Alto
    "cortexxdr": {
        "description": "Uses Palo Alto Cortex XDR field mappings",
        "pipeline": CortexXDR_pipeline(),
        "display_name": "Palo Alto Cortex XDR",
    },
    "carbonblack_enterprise": {
        "description": "Uses Carbon Black Enterprise EDR field mappings",
        "pipeline": CarbonBlack_pipeline(),
        "display_name": "CB",
    },  # Crowdstrike
    "crowdstrike_fdr": {
        "description": "Crowdstrike FDR Splunk Mappings",
        "pipeline": crowdstrike_fdr_pipeline(),
        "display_name": "CrowdStrike FDR SPL",
    },
    "crowdstrike_falcon": {
        "description": "Crowdstrike Falcon Logscale Mappings",
        "pipeline": crowdstrike_falcon_pipeline(),
        "display_name": "CrowdStrike Falcon Logscale",
    },  # Elasticsearch
    "ecs_kubernetes": {
        "description": "Elastic Common Schema (ECS) Kubernetes audit log mappings",
        "pipeline": ecs_kubernetes(),
        "display_name": "ECS Kubernetes",
    },
    "ecs_windows": {
        "description": "Elastic Common Schema (ECS) Windows log mappings from Winlogbeat from version 7",
        "pipeline": ecs_windows(),
        "display_name": "ECS Winlogbeat",
    },
    "ecs_windows_old": {
        "description": "Elastic Common Schema (ECS) Windows log mappings from Winlogbeat up to version 6",
        "pipeline": ecs_windows_old(),
        "display_name": "ESC Winlogbeat (<= v6.x)",
    },
    "ecs_zeek_beats": {
        "description": "Elastic Common Schema (ECS) for Zeek using filebeat >= 7.6.1",
        "pipeline": ecs_zeek_beats(),
        "display_name": "ECS Zeek (Elastic)",
    },
    "ecs_zeek_corelight": {
        "description": "Elastic Common Schema (ECS) mapping from Corelight",
        "pipeline": ecs_zeek_corelight(),
        "display_name": "ESC Zeek (Corelight)",
    },
    "zeek_raw": {
        "description": "Zeek raw JSON field naming",
        "pipeline": zeek_raw(),
        "display_name": "Zeek Raw JSON",
    },  # InsightIDR
    # RS uncommented this line after Stephen uncomment corresponding line in pyproject.toml
    # "insightidr": {
    #     "description": "InsightIDR Log Entry Query Language (LEQL) Transformations",
    #     "pipeline": insight_idr_pipeline(),
    #     "display_name": "InsightIDR LEQL",
    # },
    # Loki
    "loki_grafana_logfmt": {
        "description": "Converts field names to logfmt labels used by Grafana",
        "pipeline": loki_grafana_logfmt(),
        "display_name": "Logfmt Labels",
    },
    "loki_promtail_sysmon": {
        "description": "Parse and adjust field names for Windows sysmon data produced by promtail",
        "pipeline": loki_promtail_sysmon(),
        "display_name": "WinSysmon Promtail",
    },
    "loki_okta_system_log": {
        "description": "Parse the Okta System Log event json, adjusting field-names appropriately",
        "pipeline": loki_okta_system_log(),
        "display_name": "Okta System Event",
    },  # Microsoft Kusto
    "microsoft_xdr": {
        "description": "Mappings for Sysmon -> XDR Advanced Hunting Query Table Schema",
        "pipeline": microsoft_xdr_pipeline(),
        "display_name": "Microsoft XDR KustoQL",
    },  # Microsoft Sentinel ASIM
    "sentinel_asim": {
        "description": "Mappings for Sysmon -> Sentinel ASIM Query Table Schema",
        "pipeline": sentinel_asim_pipeline(),
        "display_name": "Sentinel ASIM KustoQL",
    },  # Microsoft Azure Monitor
    "azure_monitor": {
        "description": "Mappings for Sysmon -> Azure Monitor Query Table Schema",
        "pipeline": azure_monitor_pipeline(),
        "display_name": "Azure Monitor KustoQL",
    },  # Netwitness
    "netwitness_windows": {
        "description": "Netwitness Windows log mappings",
        "pipeline": netwitness_windows_pipeline(),
        "display_name": "Netwitness Windows",
    },  # QRadar
    "qradar_fields": {
        "description": "Supports only the Sigma fields in the Field Mapping",
        "pipeline": QRadarAQL_fields_pipeline(),
        "display_name": "Sigma Fields",
    },
    "qradar_payload": {
        "description": "Uses UTF8(payload) instead of fields unsupported by the Field Mapping.",
        "pipeline": QRadarAQL_payload_pipeline(),
        "display_name": "UTF8(payload) (Non-Sigma Fields)",
    },
    # Sigma Pipeline Placeholder
    "sigma_default": {
        "description": "Empty ProcessingPipeline placeholder",
        "pipeline": ProcessingPipeline(name="Sigma Placeholder"),
        "display_name": "Sigma",
    },  # SecOps
    "secops_udm": {
        "description": "Mappings for Google SecOps (Chronicle) UDM",
        "pipeline": secops_udm_pipeline(),
        "display_name": "Google SecOps UDM",
    },  # SentinelOne
    "sentinelone": {
        "description": "Mappings for SentinelOne Deep Visibility Queries",
        "pipeline": sentinelone_pipeline(),
        "display_name": "SentinelOne Deep Visibility",
    },  # Splunk
    "splunk_windows": {
        "description": "Splunk Query, Windows Mappings",
        "pipeline": splunk_windows_pipeline(),
        "display_name": "Splunk Query (Windows)",
    },
    "splunk_windows_sysmon_acc": {
        "description": "Splunk Windows Sysmon search acceleration keywords",
        "pipeline": splunk_windows_sysmon_acceleration_keywords(),
        "display_name": "Splunk Query (Sysmon)",
    },
    "splunk_cim_dm": {
        "description": "Splunk Datamodel Field Mappings",
        "pipeline": splunk_cim_data_model(),
        "display_name": "Splunk Datamodel Query",
    },  # STIX
    "stix_2_0": {"description": "STIX 2.0 Mappings", "pipeline": stix_2_0(), "display_name": "STIX 2.0"},
    "stix_shifter": {
        "description": "STIX Shifter Mappings",
        "pipeline": stix_shifter(),
        "display_name": "STIX Shifter",
    },  # Windows
    "windows_sysmon": {"description": "Sysmon for Windows", "pipeline": sysmon_pipeline(), "display_name": "Sysmon"},
    "windows_audit": {
        "description": "Windows Event Logs",
        "pipeline": windows_audit_pipeline(),
        "display_name": "Windows Event Logs",
    },
    "windows_logsource": {
        "description": "Windows Logs, General",
        "pipeline": windows_logsource_pipeline(),
        "display_name": "Windows Logs, General",
    },
}


class SigmAIQPipeline:
    """
    Implement the Factory pattern for the pySigma transformation Pipelines, which provide the functionality to
    transform SigmaRule objects. Examples are field renames, changing logsource fields, etc.

    If you would like to apply multiple processing pipelines, use the class SigmAIQPipelineResolver instead, which will
    use this factory class for each ProcessingPipeline in a list of pipelines, and return a single consolidated
    ProcessingPipeline from multiple pipelines using pySigma's ProcessingPipelineResolver class.


    """

    def __init__(self, processing_pipeline: Optional[Union[str, ProcessingPipeline, Callable]] = None):
        """Initialize the class to create a ProcessingPipeline

            :param processing_pipeline: Specifies the desired pipeline to create. This can be one of three types:
            1. A str based on the keys in PIPELINES that will load predefined pipelines according to this factory
            2. A function (or callable) that return a pySigma ProcessingPipeline object
            3. A ProcessingPipeline object
            If providing a callable or ProcessingPipeline, the ProcessingPipeline must have the 'name' attribute set
        :type processing_pipeline: str or function or ProcessingPipeline
        """

        if not processing_pipeline:  # _validate_pipeline should catch this but just to be sure
            raise ValueError("Please provide a valid processing pipeline value to processing_pipeline")
        self.pipeline = processing_pipeline

    def create_pipeline(self):
        """Simple method to just return the validated pipeline, call this after creating the class"""
        return self._validate_pipeline(self.pipeline)

    @classmethod
    def from_fieldmap(cls, fieldmapping: Dict[str, Union[str, List[str]]], name=None, priority: int = 10):
        """Creates a ProcessingPipeline object from custom field renaming dict
        :param fieldmapping: Dictionary containing field mappings in format {'original_field': 'new_field'}
        Values can be a string of the new field name, or list of strings for new field names.
        :type fieldmapping: Dict[str, Union[str, List[str]]]
        :param name: Name of the new ProcessingPipeline. If not given, defaults to
        'Custom Field Renames Pipeline <uid>'
        :type name: str, optional
        :param priority: Priority to assign to new ProcessingPipeline between 0 and 100. Pipelines with a lower
        priority will be applied first if combined with other pipelines in a resolver. Defaults to 10
        :type priority: int, optional
        """

        # Helper methods
        def _validate_field_mappings(field_mappings):
            if not field_mappings:
                raise InvalidCustomFieldMapping(
                    "The provided field_mappings is empty or None. "
                    "Please provide a valid dictionary for field mappings."
                )
            if isinstance(field_mappings, dict):
                if all(isinstance(k, str) and isinstance(v, str) for k, v in field_mappings.items()):
                    return field_mappings
            raise TypeError(
                f"fieldmappings must be of type 'dict' with str keys and values or None, "
                f"but is {type(field_mappings)}"
            )

        def _validate_name(pname):
            if pname is not None:
                if not isinstance(pname, str):
                    raise TypeError(
                        f"'name' given for new fieldmapping pipeline must be of type str, "
                        f"got {type(pname)} instead."
                    )
            return pname

        def _validate_priority(priority_num):
            """Ensure priority is valid and 0 <= priority <= 100"""
            if not priority_num:
                return 0
            if not isinstance(priority_num, int):
                raise TypeError(
                    f"custom_field_mappings_priority must be of type `int`, but {type(priority_num)} was provided"
                )
            if priority_num < 0:
                return 0
            if priority_num > 100:
                return 100
            return priority_num

        uid = str(uuid4())
        pipeline_name = _validate_name(name) or f"Custom Field Renames Pipeline {uid}"
        pipeline_priority = _validate_priority(priority)
        fieldmapping = _validate_field_mappings(fieldmapping)
        pipeline_item = ProcessingItem(identifier=uid, transformation=FieldMappingTransformation(fieldmapping))
        pipeline = ProcessingPipeline(name=pipeline_name, priority=pipeline_priority, items=[pipeline_item])
        return cls(processing_pipeline=pipeline)

    @classmethod
    def display_available_pipelines(cls) -> Dict:
        """Simple class method for returning dict of available pipelines and descriptions"""
        return {k: v.get("description") for k, v in AVAILABLE_PIPELINES.items()}

    @classmethod
    def display_available_pipelines_display_names(cls) -> Dict:
        """Simple class method for returning dict of available pipelines and display names"""
        return {k: v.get("display_name") for k, v in AVAILABLE_PIPELINES.items()}

    @staticmethod
    def _get_pipeline(pipeline_name: str) -> ProcessingPipeline:
        """Get pipeline obj from dict of PIPELINES
        :param pipeline_name: The name of the pipeline, should be a key in PIPELINES dict
        :type pipeline_name: str
        :return: pySigma ProcessingPipeline object
        :rtype: ProcessingPipeline
        """
        return AVAILABLE_PIPELINES.get(pipeline_name, {}).get("pipeline", None)

    def _validate_pipeline(self, processing_pipeline):
        """Validates if the pipeline is in the dict of available_pipelines, or is a valid ProcessingPipeline obj"""
        # If string, check to see if it's in our list of valid pipelines and call the callable
        if isinstance(processing_pipeline, str):
            if new_processing_pipeline := self._get_pipeline(processing_pipeline):
                return self._validate_pipeline(new_processing_pipeline)
            else:
                raise InvalidSigmAIQPipeline(
                    f"Provided processing_pipeline str {processing_pipeline} not in list of "
                    f"available pipelines: {', '.join(AVAILABLE_PIPELINES.keys())}"
                )

        # If it's a callable, call it and re-check it to ensure it is of type ProcessingPipeline
        if callable(processing_pipeline):
            processing_pipeline = self._validate_pipeline(processing_pipeline())
            return processing_pipeline

        # If it's a list, return SigmAIQPipelineResolver so each element/pipeline can be resolved
        if isinstance(processing_pipeline, list):
            return SigmAIQPipelineResolver(processing_pipelines=processing_pipeline).process_pipelines()

        # Last, check if is of type ProcessingPipeline
        if isinstance(processing_pipeline, ProcessingPipeline):
            return processing_pipeline

        raise TypeError(
            "Invalid processing_pipeline, or invalid object returned from callable passed as processing_pipeline"
        )


class SigmAIQPipelineResolver:
    """Takes list of processing pipelines that can be either a string, callable that returns a ProcessingPipeline,
    or ProcessingPipeline object, then consolidates them into a single pipeline based on the individual
    ProcessingPipeline's priority attribute.
    https://sigmahq-pysigma.readthedocs.io/en/latest/Processing_Pipelines.html#resolvers
    If you want to include a custom ad-hoc ProcessingPipeline in the list of supplied pipelines, create one first
    with SigmAIQPipeline().create_pipeline(), and include the returned ProcessingPipeline object in the list
    """

    def __init__(self, processing_pipelines: List[Union[str, ProcessingPipeline, Callable]]):
        """Init the class
        :param processing_pipelines: List of strings found in available_pipelines, ProcessingPipeline, or callable that
        returns a ProcessingPipeline object
        :type processing_pipelines: List[Union[str, ProcessingPipeline, Callable]]
        """
        self.processing_pipelines = self._setup(processing_pipelines)
        self.resolver = ProcessingPipelineResolver()

    @staticmethod
    def _setup(processing_pipelines):
        if processing_pipelines:
            if isinstance(processing_pipelines, list) or isinstance(processing_pipelines, set):
                return list(filter(None, processing_pipelines))
            raise TypeError(f"processing_pipelines is not of type list or set: type is {type(processing_pipelines)}")
        raise ValueError("processing_pipelines is empty or None, please provide valid values to processing_pipelines")

    def process_pipelines(self, name: Optional[str] = None) -> ProcessingPipeline:
        """Consolidates processing_pipelines with a resolver by creating a ProcessingPipeline via
        SigmAIQPipeline for each item in the processing_pipelines list.  An optional name can be passed
        to the method; if present, the final resolved singular ProcessingPipeline will be given this name. Otherwise,
        it will be named "SigmAIQ Resolved Pipelines" if multiple pipelines are in the resolver. If only one pipeline
        is in the resolver, the name will be the name of that pipeline.

        :param name: Optional, the name to give the final resolved single ProcessingPipeline. Defaults to None.
        :type name: str
        :return: Single ProcessingPipeline ordered by the priority of each pipeline returned from
        SigmAIQPipeline.
        :rtype: ProcessingPipeline
        """

        for pipeline in self.processing_pipelines:
            self.resolver.add_pipeline_class(SigmAIQPipeline(processing_pipeline=pipeline).create_pipeline())
        processing_pipeline = self.resolver.resolve(list(self.resolver.pipelines.keys()))
        list_resolver_pipelines = list(self.resolver.list_pipelines())
        num_pipelines = len(list_resolver_pipelines)  # [(name (str), ProcessingPipeline),]
        # Name final pipeline
        if name:
            processing_pipeline.name = name
        elif num_pipelines == 1:
            processing_pipeline.name = list_resolver_pipelines[0][0]
        else:
            processing_pipeline.name = "SigmAIQ Resolved Pipelines"
        return processing_pipeline



import json
from typing import List, Union

import yaml
from sigma.collection import SigmaCollection
from sigma.conversion.base import TextQueryBackend
from sigma.rule import SigmaRule

from detectiq.sigmaiq.backends.sigmaiq_abstract_backend import (
    AbstractGenericSigmAIQBackendClass,
)


class SigmAIQSigmaBackend(AbstractGenericSigmAIQBackendClass, TextQueryBackend):
    """SigmAIQ backend interface to output raw Sigma rule in various formats"""

    associated_pipelines = ["sigma_default"]
    default_pipeline = "sigma_default"
    custom_formats = {"yaml": "Default Sigma Rule output format", "json": "JSON style Sigma Rule Output"}

    # Override pySigma convert and convert_rule, since we are just outputting it as-is
    # in yaml or json format

    def translate(self, sigma_rule: Union[SigmaCollection, SigmaRule]):
        """Default implementation to translate a SigmaRule or SigmaCollection object into a Splunk query
        :param sigma_rule: SigmaRule or SigmaCollection object to translate, with or without ProcessingPipelines
        already applied to it. If a ProcessingPipeline is supplied to this class on __init__, it will automatically
        apply to the SigmaRule before translating.
        :type sigma_rule: Union[SigmaRule, SigmaCollection]
        :return: List of Splunk queries. In the case of a SigmaRule, the list will have one element.
        :rtype: list
        """
        if self.associated_pipelines and self.default_pipeline:
            self._ensure_proper_pipelines(sigma_rule)
        if isinstance(sigma_rule, SigmaRule):
            output = self._translate_rule(sigma_rule)
        else:
            output = self._translate_collection(sigma_rule)
        return output

    def convert(self, sigma_collection: SigmaCollection, **kwargs) -> List[str]:
        """Override TextBackend method. Outputs SigmaCollection object with each SigmaRule as yaml or json
        Valid output formats are as yaml or json
        :param sigma_collection: SigmaCollection of SigmaRule objects to convert to yaml or json
        :type sigma_collection: SigmaCollection
        :param **kwargs: other args
        :rtype: List[str]
        """
        output = []
        for rule in sigma_collection:
            output = output + self._translate_rule(rule)
        return output

    def convert_rule(self, sigma_rule: SigmaRule, **kwargs) -> List[str]:
        """Override TextBackend method. Outputs SigmaRule object converted to yaml or json str
        Valid output formats are as yaml or json
        :param sigma_rule: SigmaRule object to convert to yaml or json
        :type sigma_rule: SigmaRule
        :param **kwargs: other args
        :rtype: List[str]
        """
        if self.processing_pipeline:
            self.processing_pipeline.apply(sigma_rule)
        output = []
        return self.handle_output_format(sigma_rule, output)

    def handle_output_format(self, sigma_rule, output) -> List[str]:
        if self.custom_output_format == "json":
            output = [self.__to_json(sigma_rule)]
        elif self.custom_output_format == "yaml" or self.output_format == "default":
            output = [self.__to_yaml(sigma_rule)]
        return output

    @staticmethod
    def __to_json(sigma_rule) -> str:
        """Converts sigma rule to JSON str"""
        return json.dumps(sigma_rule.to_dict(), indent=2)

    @staticmethod
    def __to_yaml(sigma_rule) -> str:
        return yaml.dump(sigma_rule.to_dict(), sort_keys=False)



import logging
from copy import deepcopy
from typing import Any, Dict, List, Optional, Union

from sigma.collection import SigmaCollection
from sigma.processing.pipeline import ProcessingPipeline
from sigma.rule import SigmaRule

# Backends
from detectiq.sigmaiq.backends.carbonblack import SigmAIQCarbonBlackBackend
from detectiq.sigmaiq.backends.cortexxdr import SigmAIQCortexXDRBackend
from detectiq.sigmaiq.backends.crowdstrike import (
    SigmAIQCrowdstrikeLogscaleBackend,
    SigmAIQCrowdstrikeSplunkBackend,
)
from detectiq.sigmaiq.backends.elasticsearch import SigmAIQElasticsearchBackend
from detectiq.sigmaiq.backends.insightidr import SigmAIQInsightIDRBackend
from detectiq.sigmaiq.backends.kusto import (
    SigmAIQAzureMonitorBackend,
    SigmAIQDefenderXDRBackend,
    SigmAIQSentinelASIMBackend,
)
from detectiq.sigmaiq.backends.loki import SigmAIQLokiBackend
from detectiq.sigmaiq.backends.netwitness import SigmAIQNetwitnessBackend
from detectiq.sigmaiq.backends.opensearch import SigmAIQOpensearchBackend
from detectiq.sigmaiq.backends.qradar import SigmAIQQRadarBackend
from detectiq.sigmaiq.backends.secops import SigmAIQSecOpsBackend
from detectiq.sigmaiq.backends.sentinelone import SigmAIQSentinelOneBackend
from detectiq.sigmaiq.backends.sigma import SigmAIQSigmaBackend

## Abstract
from detectiq.sigmaiq.backends.sigmaiq_abstract_backend import (
    AbstractGenericSigmAIQBackendClass,
)
from detectiq.sigmaiq.backends.splunk import SigmAIQSplunkBackend
from detectiq.sigmaiq.backends.stix import SigmAIQStixBackend
from detectiq.sigmaiq.exceptions import InvalidSigmAIQBackend
from detectiq.sigmaiq.sigmaiq_pipeline_factory import (
    SigmAIQPipeline,
    SigmAIQPipelineResolver,
)

# Utils
from detectiq.sigmaiq.utils.sigmaiq.sigmaiq_utils import create_sigma_rule_obj

AVAILABLE_BACKENDS = {
    "carbonblack": "Carbon Black EDR",
    "cortexxdr": "Palo Alto Cortex XDR",
    "crowdstrike_splunk": "Crowdstrike FDR Splunk Query",
    "crowdstrike_logscale": "Crowdstrike Logscale Query",
    "elasticsearch": "Elastic Elasticsearch SIEM",
    # RS uncommented this line after Stephen uncomment corresponding line in pyproject.toml
    # "insightidr": "Rapid7 InsightIDR SIEM",
    "loki": "Grafana Loki LogQL SIEM",
    "microsoft_xdr": "Microsoft XDR Advanced Hunting Query (KQL) (Defender, Office365, etc)",
    "microsoft_sentinel_asim": "Microsoft Sentinel ASIM Query (KQL)",
    "microsoft_azure_monitor": "Microsoft Azure Monitor Query (KQL)",
    "netwitness": "Netwitness Query",
    "opensearch": "OpenSearch Lucene",
    "qradar": "IBM QRadar",
    "secops": "Google SecOps (Chronicle)",
    "sentinelone": "SentinelOne EDR",
    "splunk": "Splunk SIEM",
    "sigma": "Original YAML/JSON Sigma Rule Output",
    "stix": "STIX 2.0 & STIX Shifter Queries",
}


class SigmAIQBackend:
    """
    Implement the Factory pattern for the pySigma Backends, which provide the translation from SigmaRule objects
    to SIEM/Security Tool queries.
    """

    def __init__(
        self,
        backend: str,
        processing_pipeline: Optional[Union[str, list, ProcessingPipeline]] = None,
        output_format: Optional[str] = None,
    ):
        """Initialize instance attributes.

        :param backend: Specifies the desired backend.
        :type backend: str
        :param processing_pipeline: Submit processing pipeline to pySigma backend to automatically apply processing pipeline
        when converting a SigmaRule object to the backend query. This can also be done manually outside the backend,
        defaults to None
        :type processing_pipeline: ProcessingPipeline, optional
        :param output_format: Format specific to each SigmAIQBackendClass to output specific formats, such as
        'yaml' or 'json' for the raw Sigma Backend, defaults to None
        :type output_format: str, optional
        """
        logging.debug(f"Executing SigmAIQBackend constructor. backend: {backend}")
        self.backend = backend
        self.processing_pipeline = self._setup_processing_pipeline(processing_pipeline)
        self.output_format = output_format

    def create_backend(self) -> AbstractGenericSigmAIQBackendClass:
        """
        Create the factory agents.

        :return: The custom AbstractGenericSigmAIQBackendClass instance of the pySigma Backend depending on the
        value of the `backend` parameter. The abstract backend class inherits the TextBackend and specific pySigma
        backend classes attributes and methods.
        :rtype: AbstractGenericSigmAIQBackendClass
        """
        kwargs = {"processing_pipeline": self.processing_pipeline, "output_format": self.output_format}

        # Carbon Black EDR (standard & enterprise)
        if self.backend == "carbonblack":
            return SigmAIQCarbonBlackBackend(**kwargs)
        # Cortex XDR, Palo Alto
        if self.backend == "cortexxdr":
            return SigmAIQCortexXDRBackend(**kwargs)
        # Crowdstrike Splunk Query
        if self.backend == "crowdstrike_splunk":
            pipelines = ["crowdstrike_fdr", kwargs["processing_pipeline"]]
            kwargs["processing_pipeline"] = SigmAIQPipelineResolver(pipelines).process_pipelines()
            return SigmAIQCrowdstrikeSplunkBackend(**kwargs)
        # Crowdstrike Logscale Query
        if self.backend == "crowdstrike_logscale":
            pipelines = ["crowdstrike_falcon", kwargs["processing_pipeline"]]
            kwargs["processing_pipeline"] = SigmAIQPipelineResolver(pipelines).process_pipelines()
            return SigmAIQCrowdstrikeLogscaleBackend(**kwargs)
        # Elasticsearch
        if self.backend == "elasticsearch":
            return SigmAIQElasticsearchBackend(**kwargs)
        # InsightIDR
        if self.backend == "insightidr":
            return SigmAIQInsightIDRBackend(**kwargs)
        # Loki (Grafana)
        if self.backend == "loki":
            return SigmAIQLokiBackend(**kwargs)
        # Microsoft Kusto
        if self.backend == "microsoft_xdr":
            return SigmAIQDefenderXDRBackend(**kwargs)
        if self.backend == "microsoft_sentinel_asim":
            return SigmAIQSentinelASIMBackend(**kwargs)
        if self.backend == "microsoft_azure_monitor":
            return SigmAIQAzureMonitorBackend(**kwargs)
        # Netwitness
        if self.backend == "netwitness":
            return SigmAIQNetwitnessBackend(**kwargs)
        # Opensearch
        if self.backend == "opensearch":
            return SigmAIQOpensearchBackend(**kwargs)
        #
        # QRadar Backend
        if self.backend == "qradar":
            return SigmAIQQRadarBackend(**kwargs)
        # SecOps Backend
        if self.backend == "secops":
            return SigmAIQSecOpsBackend(**kwargs)
        # SentinelOne
        if self.backend == "sentinelone":
            return SigmAIQSentinelOneBackend(**kwargs)
        # Splunk Backend
        if self.backend == "splunk":
            if kwargs["output_format"] == "data_model":
                pipelines = [p for p in ["splunk_cim_dm", kwargs.get("processing_pipeline")] if p is not None]
                kwargs["processing_pipeline"] = SigmAIQPipelineResolver(pipelines).process_pipelines()
            return SigmAIQSplunkBackend(**kwargs)
        # Raw sigma output
        if self.backend == "sigma":
            if not kwargs.get("output_format"):
                kwargs["output_format"] = "yaml"
            return SigmAIQSigmaBackend(**kwargs)
        # STIX
        if self.backend == "stix":
            return SigmAIQStixBackend(**kwargs)

        raise InvalidSigmAIQBackend(
            'Backend not supported: "{}". Available backends:\n{}'.format(
                self.backend, "\n".join([f"{k}: {v}" for k, v in AVAILABLE_BACKENDS.items()])
            )
        )

    @staticmethod
    def _setup_processing_pipeline(processing_pipeline):
        """Return the processing pipeline if None or already a ProcessingPipeline
        Otherwise, try to create them with our processing_pipeline_factory"""

        if processing_pipeline:
            if not isinstance(processing_pipeline, list):
                processing_pipeline = [processing_pipeline]
        else:
            processing_pipeline = []

        # Use pipeline resolver to combine pipelines
        if processing_pipeline:
            return SigmAIQPipelineResolver(processing_pipelines=processing_pipeline).process_pipelines()
        return None

    @classmethod
    def create_all_and_translate(
        cls,
        sigma_rule: Union[SigmaRule, SigmaCollection, str, dict],
        show_errors: Optional[bool] = False,
        excluded_backends: Optional[List[str]] = None,
    ) -> Dict[Any, Any]:
        """Iterates through all combinations of backends, associated pipelines with each backend, and output formats
        for each backend, and creates a dict of outputs.

        :param sigma_rule: A valid SigmaRule or SigmaCollection object to translate
        :type sigma_rule: Union[SigmaRule, SigmaCollection]
        :param show_errors: If True, errors will be included in the list of outputs. Errors can include errors from a
        backend when specific fields cannot be converted to a query. Defaults to False
        :type show_errors: bool
        :param excluded_backends: List of backends to exclude from translations. Defaults to None
        :return: Dict of output results in the following format:
            {backend: {pipeline: {output_format: [queries]}
        :rtype: Dict[Any, Any]
        """
        backends_pipelines = cls.display_all_associated_pipelines()
        backends_output_formats = cls.display_backends_and_outputs()
        excluded_backends = [x.lower() for x in excluded_backends] if excluded_backends else []
        results = {}
        sigma_rule = create_sigma_rule_obj(sigma_rule)
        for backend, pipelines in backends_pipelines.items():
            if backend.lower() in excluded_backends:
                continue
            backend_obj = cls(backend=backend).create_backend()
            for pipeline in pipelines:
                backend_obj.set_pipeline(pipeline)
                for output_format in backends_output_formats[backend].get("output_formats"):
                    backend_obj.set_output_format(output_format)
                    output = []
                    try:
                        output = backend_obj.translate(deepcopy(sigma_rule))
                    except Exception as exc:
                        if show_errors:
                            output = [exc]
                    if output:
                        if backend not in results.keys():
                            results[backend] = {}
                        if pipeline not in results[backend].keys():
                            results[backend][pipeline] = {}
                        results[backend][pipeline][output_format] = output
        return results

    @classmethod
    def display_available_backends(cls) -> Dict[str, str]:
        """Simple class method for returning dict of available backends and descriptions
        :return: Dict containing backend name and description
        :rtype: Dict[str, str]
        """
        return AVAILABLE_BACKENDS

    @classmethod
    def display_backends_and_outputs(cls) -> Dict:
        """Display all available backends as well as their associated output formats, both standard and SigmAIQ
        custom

        :return: Dict containing available backends for keys, and a dict of output formats with descriptions as values
        :rtype: Dict
        """
        backend_formats = {}
        for backend, description in AVAILABLE_BACKENDS.items():
            backend_instance = cls(backend=backend).create_backend()
            output_formats = backend_instance.formats or {"default": "Default output format"}
            custom_output_formats = backend_instance.custom_formats or {}
            output_formats = {**output_formats, **custom_output_formats} if custom_output_formats else output_formats
            backend_formats[backend] = {}
            backend_formats[backend]["description"] = description
            backend_formats[backend]["output_formats"] = output_formats
        return backend_formats

    @classmethod
    def display_all_associated_pipelines(cls) -> Dict:
        """Displays all available backends with their associated pipelines. Output format is a dict with the available
        pipeline as a key and a dict of the pipeline name and descriptions as the value.

        :return: Dict of backend pipline information
        :rtype: Dict

        """
        associated_pipelines = {}
        available_pipelines = SigmAIQPipeline.display_available_pipelines()
        for backend in AVAILABLE_BACKENDS.keys():
            pipeline_desc = {}
            backend_instance = cls(backend=backend).create_backend()
            pipelines = backend_instance.associated_pipelines
            for pipeline in pipelines:
                pipeline_desc[pipeline] = available_pipelines[pipeline]
            associated_pipelines[backend] = pipeline_desc

        return associated_pipelines



import os


class DEFAULT_DIRS:
    ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
    SIGMA_RULE_DIR = os.path.join(ROOT_DIR, "llm/data/sigma")
    VECTOR_STORE_DIR = os.path.join(ROOT_DIR, "llm/data/vectordb")



from sigma.backends.netwitness import NetWitnessBackend

from detectiq.sigmaiq.backends.sigmaiq_abstract_backend import (
    AbstractGenericSigmAIQBackendClass,
)


class SigmAIQNetwitnessBackend(AbstractGenericSigmAIQBackendClass, NetWitnessBackend):
    """SigmAIQ backend interface for the pySigma Netwitness Backend library to translate a SigmaRule object
    to a Netwitness search query"""

    associated_pipelines = ["netwitness_windows"]
    default_pipeline = "netwitness_windows"



from abc import ABC
from typing import Dict, List, Union

from sigma.collection import SigmaCollection
from sigma.conversion.base import TextQueryBackend
from sigma.processing.pipeline import ProcessingPipeline
from sigma.rule import SigmaRule

from detectiq.sigmaiq.exceptions import InvalidOutputFormat
from detectiq.sigmaiq.sigmaiq_pipeline_factory import (
    SigmAIQPipeline,
    SigmAIQPipelineResolver,
)
from detectiq.sigmaiq.utils.sigmaiq.sigmaiq_utils import create_sigma_rule_obj


class AbstractGenericSigmAIQBackendClass(TextQueryBackend, ABC):
    """
    Abstract class for the SigmAIQBackendFactory class.

    All AIQ backends must inherit this class, as well as the specific backend class that should be used based on the
    selected pySigma backend. This way, we can still have all the functionality of a typical pySigma Backend,
    but be able to add methods and override things as needed if we need to customize things.

    For example, we add a 'custom_formats' attribute, since pySigma backends have their own dict of
    output formats defined, but we can define our own and postprocess output if we supply our own
    custom_format that doesn't appear in the dict defined in the pySigma Backend class , i.e. we want
    to take the string output from the Splunk backend, but then postprocess it and put it in a savedsearches.conf
    stanza to show to the user; this isn't available in the pySigma Splunk Backend.

    Attributes:
        custom_formats: (Dict) of any custom output formats that are not part of the parent TextBackend class 'formats'
        attribute. These should be handled by overriding the method `handle_output_format`.
        Dict should follow {'format': 'description}

        associated_pipelines: (List) of all pipelines that can/should be used with the backend. Usually, these are
        pipelines that are created alongside the backend by the backend's author. Elements in this list should be
        listed as keys in sigmaiq.sigmaiq_pipeline_factory.AVAILABLE_PIPELINES.

        default_pipeline: (str) The default pipeline name to use (as seen in AVAILABLE_PIPELINES.keys()) if no pipeline is
        provided.  Most of the time, the authors of pySigma backends will automatically apply the required pipelines,
        but some do not.
    """

    custom_formats = {}
    associated_pipelines = []
    default_pipeline = ProcessingPipeline()

    def __init__(self, processing_pipeline: ProcessingPipeline = None, output_format: str = None):
        """Initialize instance attributes.
        :param processing_pipeline: ProcessingPipeline object to apply to the SigmaRule or SigmaCollection object to
        translate. If None, no ProcessingPipelines will be applied as part of the translation process to the backend,
        but can still be manually applied outside the scope of backend translations
        :type processing_pipeline: ProcessingPipeline, optional
        :param output_format: Output format to use, typically for post-processing the SigmaRule or SigmaCollection
        translation. Each subclass implementing this class may have specific values they use to handle post-processing
        :type output_format: str
        """
        self.output_format = None
        self.custom_output_format = None
        self.set_pipeline(processing_pipeline)
        self.set_output_format(output_format)
        super().__init__(processing_pipeline=self.processing_pipeline)

    def set_pipeline(self, processing_pipeline):
        # Validate the pipeline
        self.processing_pipeline = self._validate_processing_pipeline(processing_pipeline or self.default_pipeline)
        # Ensure we aren't applying a pipeline automatically applied by the backend
        if self.backend_processing_pipeline and self.processing_pipeline:
            if self.processing_pipeline.name == self.backend_processing_pipeline.name:
                self.processing_pipeline = None

    def set_output_format(self, output_format):
        if not output_format:
            self.output_format = "default"
            self.custom_output_format = None
        else:
            self.output_format = None
            self.custom_output_format = None
            if output_format in self.custom_formats.keys():
                self.custom_output_format = output_format
            elif output_format in self.formats.keys():
                self.output_format = output_format
        if not self.output_format and not self.custom_output_format:
            raise InvalidOutputFormat(f"Invalid output_format {output_format} for Backend {type(self).__name__}")

    def translate(self, sigma_rule: Union[SigmaRule, SigmaCollection, List, Dict, str]):
        """Default implementation to translate a SigmaRule or SigmaCollection object into a Splunk query

        :param sigma_rule: SigmaRule, SigmaCollection, or valid YAML str/dictionary containing a valid Sigma rule
        to translate, with or without ProcessingPipelines already applied to it.
        If a ProcessingPipeline is supplied to this class on __init__, it will automatically apply to the SigmaRule.
        :type sigma_rule: Union[SigmaRule, SigmaCollection, List, Dict, str]
        :return: List of converted query/queries. In the case of a SigmaRule, the list will have one element.
        :rtype: list
        """
        # Ensure the sigma_rule is a SigmaRule or SigmaCollection object. If it's not, try to make one.
        sigma_rule = create_sigma_rule_obj(sigma_rule)
        # Ensure the required pipelines have been applied to the rule, or are present in the instances
        # processing pipeline attribute. Otherwise, the conversion will most likely throw errors
        if self.associated_pipelines and self.default_pipeline:
            _ = self._ensure_proper_pipelines(sigma_rule)
        # Next
        if isinstance(sigma_rule, SigmaRule):
            output = self._translate_rule(sigma_rule)
        else:
            output = self._translate_collection(sigma_rule)
        if self.custom_output_format:
            output = self.handle_output_format(sigma_rule, output)
        return output

    def handle_output_format(self, sigma_rule: Union[SigmaCollection, SigmaRule], output: List[str]):
        """Postprocessing of output of rule converted by backend for self.custom_output_format"""
        if self.custom_output_format:
            pass
        else:
            pass
        return output

    def _translate_rule(self, sigma_rule):
        return self.convert_rule(sigma_rule, output_format=self.output_format)

    def _translate_collection(self, sigma_collection):
        output = self.convert(sigma_collection, output_format=self.output_format)
        if not isinstance(output, list):
            output = [output]
        return output

    @staticmethod
    def _validate_processing_pipeline(processing_pipeline):
        if processing_pipeline and not isinstance(processing_pipeline, ProcessingPipeline):
            processing_pipeline = SigmAIQPipeline(processing_pipeline).create_pipeline()
        return processing_pipeline

    def get_backend_output_formats(self) -> Dict[str, str]:
        output_formats = {**self.formats, **self.custom_formats}
        if not output_formats:
            output_formats = {"default": "Default query string output"}
        return output_formats

    def _ensure_proper_pipelines(self, sigma_obj: Union[SigmaRule, SigmaCollection]):
        """Ensures that the default pipeline is applied to SigmaRules to be converted.
        This is to avoid potential errors when converting untransformed rules.

        It will look to see if any ProcessingItems inside ProcessingPipelines of self.associated_pipelines have been
        applied to any SigmaRule detection item objects or if any ProcessingPipelines/ProcessingItems exist in
        self.process_pipeline. If not, then self.default_pipeline will automatically be added to the instances pipeline
        and applied at rule conversion time.

        :param sigma_obj:
        :type sigma_obj: Union[SigmaRule, SigmaCollection]
        """
        associated_pipelines = [
            SigmAIQPipeline(processing_pipeline=p).create_pipeline() for p in self.associated_pipelines
        ]
        set_processing_pipeline_item_ids = (
            [x.identifier for x in self.processing_pipeline.items] if self.processing_pipeline else []
        )

        # Check if self.processing_pipeline has the same name as any associated pipelines
        if self.processing_pipeline:
            if any(x.name == self.processing_pipeline.name for x in associated_pipelines):
                return True

        # Maybe pipelines were combined, so check if all ProcessingItems from any associated pipeline exist
        for associated_pipeline in associated_pipelines:
            if all(proc_item.identifier in set_processing_pipeline_item_ids for proc_item in associated_pipeline.items):
                return True
            # Check to see if it's automatically applied by the backend
            if associated_pipeline.name == self.backend_processing_pipeline.name:
                return True
            if isinstance(sigma_obj, SigmaRule):
                if any(
                    proc_item.identifier in sigma_obj.applied_processing_items
                    for proc_item in associated_pipeline.items
                ):
                    return True
            else:
                for sigma_rule in sigma_obj:
                    if any(
                        proc_item.identifier in sigma_rule.applied_processing_items
                        for proc_item in associated_pipeline.items
                    ):
                        return True

        if not self.processing_pipeline:  # No pipeline currently
            self.processing_pipeline = SigmAIQPipeline(processing_pipeline=self.default_pipeline).create_pipeline()
        else:  # We have a pipeline, add the default to the beginning of it
            self.processing_pipeline = SigmAIQPipelineResolver(
                processing_pipelines=[self.default_pipeline, self.processing_pipeline]
            ).process_pipelines()
